/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 *
 * This file is automatically generated by buildiface
 * DO NOT EDIT
 */

#include "cdesc.h"

int MPIR_Get_address_cdesc(CFI_cdesc_t* x0, MPI_Aint * x1)
{
    int err = MPI_SUCCESS;
    void *buf0 = x0->base_addr;

    if (buf0 == &MPIR_F08_MPI_BOTTOM) {
        buf0 = MPI_BOTTOM;
    }

    err = MPI_Get_address(buf0, x1);

    return err;
}
