/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef POSIX_EAGER_IQUEUE_DIRECT_H_INCLUDED
#define POSIX_EAGER_IQUEUE_DIRECT_H_INCLUDED

#include "iqueue_noinline.h"
#include "iqueue_send.h"
#include "iqueue_recv.h"

#endif /* POSIX_EAGER_IQUEUE_DIRECT_H_INCLUDED */
