/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef HCOLLPRE_H_INCLUDED
#define HCOLLPRE_H_INCLUDED

typedef struct {
    int is_hcoll_init;
    void *hcoll_context;
} hcoll_comm_priv_t;

#endif /* HCOLLPRE_H_INCLUDED */
