!
! Copyright (C) by Argonne National Laboratory
!     See COPYRIGHT in top-level directory
!

module mpi_c_interface
    use mpi_c_interface_types
    use mpi_c_interface_nobuf
    use mpi_c_interface_cdesc
    use mpi_c_interface_glue
end module mpi_c_interface
