/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef SHM_INLINE_H_INCLUDED
#define SHM_INLINE_H_INCLUDED

#include "stubshm_init.h"
#include "stubshm_probe.h"
#include "stubshm_recv.h"
#include "stubshm_request.h"
#include "stubshm_send.h"
#include "stubshm_win.h"
#include "stubshm_rma.h"
#include "stubshm_am.h"
#include "stubshm_spawn.h"
#include "stubshm_comm.h"
#include "stubshm_coll.h"
#include "stubshm_unimpl.h"

#endif /* SHM_INLINE_H_INCLUDED */
