/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef RSH_H_INCLUDED
#define RSH_H_INCLUDED

#include "hydra.h"

HYD_status HYDT_bscd_rsh_query_env_inherit(const char *env_name, int *ret);

#endif /* RSH_H_INCLUDED */
