#ifndef LOCK_INTERNAL_H_INCLUDED
#define LOCK_INTERNAL_H_INCLUDED

#include <fcntl.h>

const char *ADIOI_GEN_flock_cmd_to_string(int cmd);
const char *ADIOI_GEN_flock_type_to_string(int type);

#endif /* LOCK_INTERNAL_H_INCLUDED */
