/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef MPIDI_NEM_STATISTICS_H_INCLUDED
#define MPIDI_NEM_STATISTICS_H_INCLUDED

extern unsigned long long *MPID_nem_fbox_fall_back_to_queue_count;

#endif /* MPIDI_NEM_STATISTICS_H_INCLUDED */
