/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef HYDRA_TOPO_INTERNAL_H_INCLUDED
#define HYDRA_TOPO_INTERNAL_H_INCLUDED

#include "hydra_topo.h"

struct HYDI_topo_info {
    int debug;
};

extern struct HYDI_topo_info HYDI_topo_info;

#endif /* HYDRA_TOPO_INTERNAL_H_INCLUDED */
