/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef MPID_NEM_ATOMICS_H_INCLUDED
#define MPID_NEM_ATOMICS_H_INCLUDED

#include "mpichconf.h"
#include "mpl.h"

#endif /* MPID_NEM_ATOMICS_H_INCLUDED */
