/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 *
 * This file is automatically generated by buildiface 
 * DO NOT EDIT
 */
#include "mpi_fortimpl.h"


/* Begin MPI profiling block */
#if defined(USE_WEAK_SYMBOLS) && !defined(USE_ONLY_MPI_NAMES) 
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak MPI_FILE_GET_AMODE = PMPI_FILE_GET_AMODE
#pragma weak mpi_file_get_amode__ = PMPI_FILE_GET_AMODE
#pragma weak mpi_file_get_amode_ = PMPI_FILE_GET_AMODE
#pragma weak mpi_file_get_amode = PMPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_FILE_GET_AMODE = pmpi_file_get_amode__
#pragma weak mpi_file_get_amode__ = pmpi_file_get_amode__
#pragma weak mpi_file_get_amode_ = pmpi_file_get_amode__
#pragma weak mpi_file_get_amode = pmpi_file_get_amode__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_FILE_GET_AMODE = pmpi_file_get_amode_
#pragma weak mpi_file_get_amode__ = pmpi_file_get_amode_
#pragma weak mpi_file_get_amode_ = pmpi_file_get_amode_
#pragma weak mpi_file_get_amode = pmpi_file_get_amode_
#else
#pragma weak MPI_FILE_GET_AMODE = pmpi_file_get_amode
#pragma weak mpi_file_get_amode__ = pmpi_file_get_amode
#pragma weak mpi_file_get_amode_ = pmpi_file_get_amode
#pragma weak mpi_file_get_amode = pmpi_file_get_amode
#endif



#elif defined(HAVE_PRAGMA_WEAK)

#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak MPI_FILE_GET_AMODE = PMPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_file_get_amode__ = pmpi_file_get_amode__
#elif !defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_file_get_amode = pmpi_file_get_amode
#else
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#pragma weak mpi_file_get_amode_ = pmpi_file_get_amode_
#endif

#elif defined(HAVE_PRAGMA_HP_SEC_DEF)
#if defined(F77_NAME_UPPER)
#pragma _HP_SECONDARY_DEF PMPI_FILE_GET_AMODE  MPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _HP_SECONDARY_DEF pmpi_file_get_amode__  mpi_file_get_amode__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _HP_SECONDARY_DEF pmpi_file_get_amode  mpi_file_get_amode
#else
#pragma _HP_SECONDARY_DEF pmpi_file_get_amode_  mpi_file_get_amode_
#endif

#elif defined(HAVE_PRAGMA_CRI_DUP)
#if defined(F77_NAME_UPPER)
#pragma _CRI duplicate MPI_FILE_GET_AMODE as PMPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma _CRI duplicate mpi_file_get_amode__ as pmpi_file_get_amode__
#elif !defined(F77_NAME_LOWER_USCORE)
#pragma _CRI duplicate mpi_file_get_amode as pmpi_file_get_amode
#else
#pragma _CRI duplicate mpi_file_get_amode_ as pmpi_file_get_amode_
#endif

#elif defined(HAVE_WEAK_ATTRIBUTE)
#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));

#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));

#elif defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));

#else
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));

#endif
#endif /* HAVE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */
/* End MPI profiling block */


/* These definitions are used only for generating the Fortran wrappers */
#if defined(USE_WEAK_SYMBOLS) && defined(USE_ONLY_MPI_NAMES)
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#if defined(F77_NAME_UPPER)
#pragma weak mpi_file_get_amode__ = MPI_FILE_GET_AMODE
#pragma weak mpi_file_get_amode_ = MPI_FILE_GET_AMODE
#pragma weak mpi_file_get_amode = MPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak MPI_FILE_GET_AMODE = mpi_file_get_amode__
#pragma weak mpi_file_get_amode_ = mpi_file_get_amode__
#pragma weak mpi_file_get_amode = mpi_file_get_amode__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak MPI_FILE_GET_AMODE = mpi_file_get_amode_
#pragma weak mpi_file_get_amode__ = mpi_file_get_amode_
#pragma weak mpi_file_get_amode = mpi_file_get_amode_
#else
#pragma weak MPI_FILE_GET_AMODE = mpi_file_get_amode
#pragma weak mpi_file_get_amode__ = mpi_file_get_amode
#pragma weak mpi_file_get_amode_ = mpi_file_get_amode
#endif
#elif defined(HAVE_WEAK_ATTRIBUTE)
#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("MPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("MPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("MPI_FILE_GET_AMODE")));

#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode__")));

#elif defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * );
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode_")));

#else
extern FORT_DLL_SPEC void FORT_CALL MPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("mpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif
#endif

#endif

/* Map the name to the correct form */
#ifndef MPICH_MPI_FROM_PMPI
#if defined(USE_WEAK_SYMBOLS)
#if defined(HAVE_MULTIPLE_PRAGMA_WEAK)
/* Define the weak versions of the PMPI routine*/
#ifndef F77_NAME_UPPER
extern FORT_DLL_SPEC void FORT_CALL PMPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_2USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER_USCORE
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * );
#endif
#ifndef F77_NAME_LOWER
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * );

#endif

#if defined(F77_NAME_UPPER)
#pragma weak pmpi_file_get_amode__ = PMPI_FILE_GET_AMODE
#pragma weak pmpi_file_get_amode_ = PMPI_FILE_GET_AMODE
#pragma weak pmpi_file_get_amode = PMPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#pragma weak PMPI_FILE_GET_AMODE = pmpi_file_get_amode__
#pragma weak pmpi_file_get_amode_ = pmpi_file_get_amode__
#pragma weak pmpi_file_get_amode = pmpi_file_get_amode__
#elif defined(F77_NAME_LOWER_USCORE)
#pragma weak PMPI_FILE_GET_AMODE = pmpi_file_get_amode_
#pragma weak pmpi_file_get_amode__ = pmpi_file_get_amode_
#pragma weak pmpi_file_get_amode = pmpi_file_get_amode_
#else
#pragma weak PMPI_FILE_GET_AMODE = pmpi_file_get_amode
#pragma weak pmpi_file_get_amode__ = pmpi_file_get_amode
#pragma weak pmpi_file_get_amode_ = pmpi_file_get_amode
#endif /* Test on name mapping */

#elif defined(HAVE_WEAK_ATTRIBUTE)
#if defined(F77_NAME_UPPER)
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("PMPI_FILE_GET_AMODE")));

#elif defined(F77_NAME_LOWER_2USCORE)
extern FORT_DLL_SPEC void FORT_CALL PMPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode__")));

#elif defined(F77_NAME_LOWER_USCORE)
extern FORT_DLL_SPEC void FORT_CALL PMPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode_")));

#else
extern FORT_DLL_SPEC void FORT_CALL PMPI_FILE_GET_AMODE( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode__( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));
extern FORT_DLL_SPEC void FORT_CALL pmpi_file_get_amode_( MPI_Fint *, MPI_Fint *, MPI_Fint * ) __attribute__((weak,alias("pmpi_file_get_amode")));

#endif /* Test on name mapping */
#endif /* HAVE_MULTIPLE_PRAGMA_WEAK */
#endif /* USE_WEAK_SYMBOLS */

#ifdef F77_NAME_UPPER
#define mpi_file_get_amode_ PMPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_file_get_amode_ pmpi_file_get_amode__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_file_get_amode_ pmpi_file_get_amode
#else
#define mpi_file_get_amode_ pmpi_file_get_amode_
#endif /* Test on name mapping */

#ifdef F77_USE_PMPI
/* This defines the routine that we call, which must be the PMPI version
   since we're renaming the Fortran entry as the pmpi version.  The MPI name
   must be undefined first to prevent any conflicts with previous renamings. */
#undef MPI_File_get_amode
#define MPI_File_get_amode PMPI_File_get_amode 
#endif

#else

#ifdef F77_NAME_UPPER
#define mpi_file_get_amode_ MPI_FILE_GET_AMODE
#elif defined(F77_NAME_LOWER_2USCORE)
#define mpi_file_get_amode_ mpi_file_get_amode__
#elif !defined(F77_NAME_LOWER_USCORE)
#define mpi_file_get_amode_ mpi_file_get_amode
/* Else leave name alone */
#endif


#endif /* MPICH_MPI_FROM_PMPI */

/* Prototypes for the Fortran interfaces */
#include "fproto.h"
FORT_DLL_SPEC void FORT_CALL mpi_file_get_amode_ ( MPI_Fint *v1, MPI_Fint *v2, MPI_Fint *ierr ){
#ifdef MPI_MODE_RDONLY
    *ierr = MPI_File_get_amode( MPI_File_f2c(*v1), v2 );
#else
*ierr = MPI_ERR_INTERN;
#endif
}
